/**
 * File: cdemu.c
 * Authors: Robert Penz <robert.penz@outertech.com>
 *          Mike Frysinger <vapier@gmail.com>
 * Copyright 2002-2004
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the Free Software           *
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.             *
 *                                                                         *
 ***************************************************************************/

#include "cdemu_kernel.h"
#include "cdemu.h"

static inline long cdemu_getoffset(struct cdemu_device *vc, int track, int lba);
static inline void lba_to_msf(int lba, unsigned short *m, unsigned short *s, unsigned short *f);
static inline int msf_to_lba(int m, int s, int f);
static inline int cdemu_mode_2_sector_size(int mode);
static void convert_track_mode(struct cdrom_tocentry *ep, int mode);
static int cdemu_findtrack(struct cdemu_device *vc, int sector);
static int read_toc_header(struct cdrom_tochdr *hp, struct cdemu_device *vc);
static void get_toc_entry(struct cdrom_tocentry *ep, struct cdemu_device *vc);
static int cdemu_read_actor_bio(read_descriptor_t *desc, struct page *page,
                                unsigned long offset, unsigned long size);
static int cdemu_transfer_bio(struct cdemu_device *vc, struct bio *bio, sector_t sector);
static int unload_cd(struct cdemu_device *vc);

/* *********************************************************************** */
/*                        helper functions (generic kernel)                */
/* *********************************************************************** */

static inline long cdemu_getoffset(struct cdemu_device *vc, int track, int lba)
{
	return vc->tracks[track].start_offset + 
	(lba - vc->tracks[track].start_sector) * cdemu_mode_2_sector_size(vc->tracks[track].mode);
}

static inline void lba_to_msf(int lba, unsigned short *m, unsigned short *s, unsigned short *f)
{
	lba += CD_MSF_OFFSET;
	lba &= 0xffffff;  /* negative lbas use only 24 bits */
	*m = lba / (CD_SECS * CD_FRAMES);
	lba %= (CD_SECS * CD_FRAMES);
	*s = lba / CD_FRAMES;
	*f = lba % CD_FRAMES;
}


static inline int msf_to_lba(int m, int s, int f)
{
	return (((m * CD_SECS) + s) * CD_FRAMES + f) - CD_MSF_OFFSET;
}

static inline int cdemu_mode_2_sector_size(int mode)
{
	switch (mode) {
		case AUDIO:		return CD_FRAMESIZE_RAW;
		case AUDIO_2448:	return CD_FRAMESIZE_RAW+CD_FRAMESIZE_SUB;
		case MODE1_2448:	return CD_FRAMESIZE_RAW+CD_FRAMESIZE_SUB;
		case MODE1_2352:	return CD_FRAMESIZE_RAW;
		case MODE1_2048:	return CD_FRAMESIZE;
		case MODE2_2352:	return CD_FRAMESIZE_RAW;
		case MODE2_2336:	return CD_FRAMESIZE_RAW0;
		default:		return 0;
	}
}

static void convert_track_mode(struct cdrom_tocentry *ep, int mode)
{
	ep->cdte_datamode = 0;
	switch (mode) {
        case AUDIO_2448:
		case AUDIO:			ep->cdte_ctrl = 0;
							break;
        case MODE1_2448:
		case MODE1_2352:
		case MODE1_2048:
		case MODE2_2352:
		case MODE2_2336:	ep->cdte_ctrl = ep->cdte_ctrl | CDROM_DATA_TRACK;
							break;
	}
	return;
}

static int cdemu_findtrack(struct cdemu_device *vc, int sector)
{
	int i;
	for (i = vc->num_tracks - 1; i >= 0; i--)
		if (vc->tracks[i].start_sector <= sector)
			return i;
	return 0;
}

/* return 0 upon success */
static int read_toc_header(struct cdrom_tochdr *hp, struct cdemu_device *vc)
{
	if (vc->loaded == 0)
		dprintk("read_toc_header: here is no cd loaded.\n");
	if (hp) {
		hp->cdth_trk0 = 1;
		hp->cdth_trk1 = vc->num_tracks ;
		return 0;
	}
	return -1;
}

/* cdrom.c guarantees that cdte_format == CDROM_MSF */
static void get_toc_entry(struct cdrom_tocentry *ep, struct cdemu_device *vc)
{
	tTrack *aTrack;

	if (vc->loaded == 0)
		dprintk("get_toc_entry: here is no cd loaded\n");

	/* lead out or in our boundies, everthing els is an erro*/
	if (ep->cdte_track == CDROM_LEADOUT)
		aTrack = &(vc->tracks[vc->num_tracks]);
	else if (ep->cdte_track > vc->num_tracks  || ep->cdte_track < 1) {
		dprintk("get_toc_entry: there is no track %i\n", ep->cdte_track);
		return;
	} else /* normal case */
		aTrack = &(vc->tracks[ep->cdte_track]) - 1;

	ep->cdte_addr.msf.frame = aTrack->frame;
	ep->cdte_addr.msf.second = aTrack->second;
	ep->cdte_addr.msf.minute = aTrack->minute;

	convert_track_mode(ep, aTrack->mode);

	dprintk("get_toc_entry: track %i, msf %i:%i%i, mode %i\n",
	        ep->cdte_track,
	        aTrack->minute, aTrack->second, aTrack->frame,
	        aTrack->mode);
}

static int cdemu_read_actor_bio(read_descriptor_t *desc, struct page *page,
                                unsigned long offset, unsigned long size)
{
	struct bio *bio = desc->arg.data;
	char *src_buf = kmap_atomic(page, KM_USER0) + offset;
	char *dst_buf = kmap_atomic(bio_page(bio), KM_USER1) + bio_offset(bio);

	if (size > desc->count - desc->written)
		size = desc->count - desc->written;

	memcpy(dst_buf + desc->written, src_buf, size);
	desc->written += size;

	kunmap_atomic(src_buf, KM_USER0);
	kunmap_atomic(dst_buf, KM_USER1);
	cond_resched();

	return size;
}

static int cdemu_transfer_bio(struct cdemu_device *vc, struct bio *bio, sector_t sector)
{
	ssize_t retval;
	int startposition, track;
	int offset;
	loff_t position;
	/* is only != 0 if the sector is not inline with our sectors
	vmware does request such sectors ... mount does not */
	int sec_offset = sector % 4;
	sector /= 4;

	/* thats just testing code as I don't know really how it works */

	track = cdemu_findtrack(vc, sector);

	/* we know the track get the mode of it */
	switch (vc->tracks[track].mode) {
		case MODE1_2352:
		case MODE1_2448:
			offset = CD_HEAD_SIZE + CD_SYNC_SIZE;
			break;
		case MODE1_2048:
			offset = 0;
			break;
		case MODE2_2352:
			offset = CD_SYNC_SIZE + CD_HEAD_SIZE + CD_SUBHEAD_SIZE;
			break;
		case MODE2_2336:
			offset = CD_SUBHEAD_SIZE;
			break;
		default:
			if (printk_ratelimit())
				printk_cdemu(KERN_WARNING, "Normal read not possible for track mode '%i'\n", vc->tracks[track].mode);
			return 0;
	}

	/* get the mode of the bin file part and calc the positon */
	startposition = cdemu_getoffset(vc, track, sector);

	/* we only support read, and we told the kernel that, 
	 * so we won't check it here */
	position = startposition + offset + sec_offset * 512;
	dprintk("cdemu_transfer: %lli %i\n", position, bio_iovec(bio)->bv_len);
	retval = vc->backing_file->f_op->sendfile(vc->backing_file,
		&position, bio_iovec(bio)->bv_len, cdemu_read_actor_bio, bio);
	if (retval < 0) {
		dprintk("cdemu_transfer: retval error %i\n", (int) retval);
		return retval;
	} else {
		return 0;
	}
}

static int unload_cd(struct cdemu_device *vc)
{
	dprintk("unload_cd: unlocking start\n");

	/* check if the cdrom is locked */
	if (vc->locked)
		return -EBUSY;

	/* make sure uid is ok */
	if (current->uid != 0 && vc->uid != current->uid)
		return -EACCES;

	/* just close the bin file */
	set_capacity(vc->gd, 0);
	vc->loaded = 0;
	vc->size = 0;
	vc->num_tracks = 0;
	strncpy(vc->comment, "", CDEMU_COMMENT_SIZE);
	fput(vc->backing_file);
	DOWN_MODULE_USE();

	dprintk("unload_cd: unlocking end\n");
	return 0;
}

/* *********************************************************************** */
/*                        helper functions (specific kernel)               */
/* *********************************************************************** */

/* *********************************************************************** */
/*                        device stuff                                     */
/* *********************************************************************** */

extern int cdemu_make_request(request_queue_t *q, struct bio *bio)
{
	struct cdemu_device *vc;
	struct bio_vec *bvec;
	int status = 0;
	int seqno;
	sector_t sector = bio->bi_sector;

	/* Which "device" are we using? */
	vc = bio->bi_bdev->bd_disk->private_data;
	if (vc == NULL) {
		if (printk_ratelimit())
			printk_cdemu(KERN_WARNING, "Request for non-existant device\n");
		status = -ENODEV;
		goto end_this_request;
	}
	/* check if we have a "cd loaded" */
	if (!vc->loaded) {
		printk_cdemu(KERN_WARNING, "Request when drive #%i isn't loaded\n", vc->number);
		status = -ENOMEDIUM;
		goto end_this_request;
	}

	/* Perform the transfers, device seems to be good */
	bio_for_each_segment(bvec, bio, seqno) {
		status = cdemu_transfer_bio(vc, bio, sector);
		if (status < 0)
			break;
		sector += bio_cur_sectors(bio);
	}

	/* now lets end the request */
end_this_request:
	bio_endio(bio, bio->bi_size, status);
	return 0;
}

/*                        interface to the user space */
static int cdemu_open(struct cdrom_device_info *cdi, int purpose)
{
	struct cdemu_device *vc;

	dprintk("cdemu_open: start %s\n", cdi->name);

	/* lets find the device and call it vc for easier use */
	if (!cdi || !cdi->handle) {
		dprintk("cdemu_open: -ENODEV\n");
		return -ENODEV;
	}
	vc = cdi->handle;

	dprintk("cdemu_open: end\n");
	return 0;
}

/*                        interface to the user space */
static void cdemu_release(struct cdrom_device_info *cdi)
{
	struct cdemu_device *vc;

	dprintk("cdemu_release: start\n");

	/* lets find the device and call it vc for easier use */
	if (!cdi || !cdi->handle)
		return;
	vc = cdi->handle;

	dprintk("cdemu_release: end\n");
	return;
}

static int cdemu_drive_status(struct cdrom_device_info *cdi, int slot_nr)
{
	struct cdemu_device *vc;

	dprintk("cdemu_drive_status: start\n");

	/* lets find the device and call it vc for easier use */
	if (!cdi || !cdi->handle) {
		dprintk("cdemu_drive_status: -ENODEV\n");
		return -ENODEV;
	}
	vc = cdi->handle;

	dprintk("cdemu_drive_status: loaded: %i\n", vc->loaded);
	return (vc->loaded) ? CDS_DISC_OK : CDS_NO_DISC;
}

static int cdemu_media_changed(struct cdrom_device_info *cdi, int disc_nr)
{
	struct cdemu_device *vc;
	int ret;

	dprintk("cdemu_media_changed: start %s\n", cdi->name);

	/* lets find the device and call it vc for easier use */
	if (!cdi || !cdi->handle) {
		dprintk("cdemu_media_changed: -ENODEV\n");
		return -ENODEV;
	}
	vc = cdi->handle;

	ret = vc->media_changed;
	vc->media_changed = 0;  /* clear bit */

	dprintk("cdemu_media_changed: end\n");
	return ret;
}

static int cdemu_tray_move(struct cdrom_device_info *cdi, int position)
{
	struct cdemu_device *vc;

	dprintk("cdemu_tray_move: start\n");

	/* lets find the device and call it vc for easier use */
	if (!cdi || !cdi->handle)
		return -ENODEV;
	vc = cdi->handle;

	/* 1: eject */
	if (position)
		if (vc->loaded)
			unload_cd(vc);

	dprintk("cdemu_tray_move: end\n");
	return 0;
}

/* locks or unlocks door lock==1: lock; return 0 upon success */
static int cdemu_lock_door(struct cdrom_device_info *cdi, int lock)
{
	struct cdemu_device *vc;

	dprintk("cdemu_lock_door: start\n");

	/* lets find the device and call it vc for easier use */
	if (!cdi || !cdi->handle)
		return -ENODEV;
	vc = cdi->handle;

	vc->locked = lock;

	return 0;
}

/* no idea how I should implement that ... so just say we set it */
static int cdemu_select_speed(struct cdrom_device_info *cdi, int speed)
{
	dprintk("cdemu_select_speed: start\n");
	return 1;
}

/* bin/cue files can not be in mutlisession */
static int cdemu_get_last_session(struct cdrom_device_info *cdi, 
                                  struct cdrom_multisession *mssp)
{
	dprintk("cdemu_get_last_session: start\n");

	/* the cd is not multi session */
	mssp->xa_flag = 0;

	return 0;
}

static int cdemu_reset_helper(struct device *dev, void *ptr)
{
	struct cdemu_device *vc;
	
	vc = to_cdemu_device(dev);
	if(!vc)
		return -1;

	unload_cd(vc);
	
	return 0;
}

static int cdemu_reset(struct cdrom_device_info *cdi)
{
	dprintk("cdemu_reset: start\n");

	device_for_each_child(&virtual_bus, NULL, &cdemu_reset_helper);

	return 0;
}

static int cdemu_audio_ioctl(struct cdrom_device_info *cdi, 
                             unsigned int cmd, void *arg)
{
	struct cdemu_device *vc;

	dprintk("cdemu_audio_ioctl: start\n");

	/* lets find the device and call it vc for easier use */
	if (!cdi || !cdi->handle) {
		dprintk("cdemu_audio_ioctl: -ENODEV\n");
		return -ENODEV;
	}
	vc = cdi->handle;

	switch (cmd) {
		case CDROMREADTOCHDR:
			return read_toc_header((struct cdrom_tochdr *) arg, vc);
		case CDROMREADTOCENTRY:
			get_toc_entry((struct cdrom_tocentry *) arg, vc);
			return 0;
		case CDROMPLAYMSF:
		case CDROMPLAYTRKIND:
		case CDROMSTOP:
		case CDROMPAUSE:
		case CDROMRESUME:
		case CDROMSTART:
		case CDROMVOLCTRL:
			return 0;
		case CDROMSUBCHNL:
		default:
			return -EINVAL;
	}
}

/* RENE: no idea how I should implement that ... so just print that command */
static int cdemu_generic_packet(struct cdrom_device_info *cdi, 
                                struct packet_command *pc)
{
	int ret = -EIO;
	int i;
	dprintk("cdemu_generic_packet: start -");
	for (i = 0; i < CDROM_PACKET_SIZE; i++)
		printk(" %.2x", pc->cmd[i]);
	printk("\n");

	if (pc->sense) {
		dprintk("cdemu_generic_packet: sense unknown\n");
		pc->sense->sense_key = 0x05;
		pc->sense->asc = 0x20;
		pc->sense->ascq = 0x00;
	}

	pc->stat = ret;
	return ret;	
}

static int cdemu_read_data(struct cdrom_device_info *cdi, __u8 __user *ubuf, int lba, int nframes, int mode)
{
	struct cdemu_device *vc = cdi->handle;
	ssize_t delta = -1, retval = 0;
	int t, frame_size;
	size_t size;
	loff_t off;
	
	t = cdemu_findtrack(vc, lba);
	off = cdemu_getoffset(vc, t, lba);
	frame_size = cdemu_mode_2_sector_size(mode);
	size = nframes * frame_size;

	dprintk("cdemu_read_data: reading %d frame(s) starting at lba %d\n", nframes, lba);

	if (vc->tracks[t].mode == mode)
		delta = 0;
	else
	{
		switch (vc->tracks[t].mode)
		{
		case AUDIO_2448:
			if (mode == AUDIO)
				delta = CD_FRAMESIZE_SUB;
			break;
		case MODE1_2448:
			if (mode == MODE1_2352)
				delta = CD_FRAMESIZE_SUB;
			break;
		}
	}


	if (delta == 0)
		retval = vc->backing_file->f_op->read(vc->backing_file, ubuf, size, &off);
	else if (delta > 0)
	{
		dprintk("cdemu_read_data: scatter-gather %d frames\n", nframes);

		while (nframes-- > 0)
		{
			ssize_t read = vc->backing_file->f_op->read(vc->backing_file, ubuf, frame_size, &off);

			if (read != frame_size)
				break;

			retval += read;
			off += delta;
			ubuf += frame_size;
		}
	}
	else
		dprintk("cdemu_read_data: source mode %d, but target mode %d!\n", vc->tracks[t].mode, mode);

	if (retval != size)
	{
		dprintk("cdemu_read_data: only %i bytes read\n", retval);

		return -EIO;
	}

	return 0;
}

static int cdemu_ioctl(struct cdrom_device_info *cdi, unsigned int cmd, 
                       unsigned long arg)
{
	struct cdemu_device *vc;
	struct cdrom_msf msf;
	struct cdemu_info cdemu_info;

	sector_t gd_sector;
	loff_t   position;
	int      sector;
	int      track;
	int      tmp;

	dprintk("cdemu_ioctl: start\n");

	/* lets find the device and call it vc for easier use */
	if (!cdi || !cdi->handle)
		return -ENODEV;
	vc = cdi->handle;

	if (cmd != CDROMREADRAW)
		dprintk("cdemu_ioctl: start - command:%x at cdrom %i\n", cmd, vc->number);

	/* check if we've a "cd loaded" and we're not trying to load one */
	if (vc->loaded == 0 && cmd != CDEMU_LOADCD) {
		if (cmd != CDROMREADRAW)
			dprintk("cdemu_ioctl: here is no cd loaded.\n");
		return -ENOMEDIUM;
	}

	switch (cmd) {
		case CDROMREADCOOKED:	/*read data in mode 1 (2048 Bytes) */
		case CDROMREADRAW:		/*read data in mode 2 (2336 Bytes) */
			if (copy_from_user(&msf, (void *) arg, sizeof(msf)))
				return -EFAULT;

			sector = msf_to_lba(msf.cdmsf_min0, msf.cdmsf_sec0, msf.cdmsf_frame0);
			track = cdemu_findtrack(vc, sector);

			/* get the mode of the bin file part and calc the positon */
			position = vc->tracks[track].start_offset + 
					(sector - vc->tracks[track].start_sector)
					* cdemu_mode_2_sector_size(vc->tracks[track].mode);

			dprintk("sector:%i track:%i position %i (%i:%i:%i)\n",
			        sector, track, (int)position,
			        msf.cdmsf_min0, msf.cdmsf_sec0, msf.cdmsf_frame0);

			/* check if the track is at its end*/
			if (position >= vc->tracks[track+1].start_offset) {
				dprintk("cdemu_ioctl: position out of track - sector:%i track:%i position %i\n",
						sector, track, (int)position);
				return -EFAULT;
			}

			if (cmd == CDROMREADRAW) {
				struct iovec local_iov = { .iov_base = (void*)arg, .iov_len = CD_FRAMESIZE_RAW };
				struct kiocb kiocb;
				
				init_sync_kiocb(&kiocb, vc->backing_file);
				tmp = generic_file_aio_read(&kiocb, &local_iov, 1, position);
				if (-EIOCBQUEUED == tmp)
					tmp = wait_on_sync_kiocb(&kiocb);      

				if (tmp != CD_FRAMESIZE_RAW)
						dprintk("only read %i bytes\n",tmp);
			} else {	/*CDROMREADCOOKED*/
				struct iovec local_iov = { .iov_base = (void*)arg, .iov_len = CD_FRAMESIZE };
				struct kiocb kiocb;
				
				init_sync_kiocb(&kiocb, vc->backing_file);
				tmp = generic_file_aio_read(&kiocb, &local_iov, 1, position);
				if (-EIOCBQUEUED == tmp)
					tmp = wait_on_sync_kiocb(&kiocb);

				if (tmp != CD_FRAMESIZE)
					dprintk("only read %i bytes\n",tmp);
			}

			return 0;
		case CDROMREADAUDIO:
			{
			struct cdrom_read_audio ra;
			int lba;

			if (copy_from_user(&ra, (struct cdrom_read_audio *) arg, sizeof(ra)))
				return -EFAULT;

			if (ra.addr_format == CDROM_MSF)
				lba = msf_to_lba(ra.addr.msf.minute,
						ra.addr.msf.second,
						ra.addr.msf.frame);
			else if (ra.addr_format == CDROM_LBA)
				lba = ra.addr.lba;
			else
				return -EINVAL;

			if (lba < 0 || ra.nframes <= 0 || ra.nframes > CD_FRAMES)
				return -EINVAL;

			return cdemu_read_data(cdi, ra.buf, lba, ra.nframes, AUDIO);
			}
		case CDROMSEEK:			/* seek msf address - not supported */
			return 0;

		case CDROMREADMODE1:	/* set read data in mode 1 */
		case CDROMREADMODE2:	/* set read data in mode 2 */
			dprintk("cdemu_ioctl: not supported that CDROMREAD: %x\n", cmd);
			return 0;

		case CDEMU_LOADCD:
			dprintk("cdemu_ioctl: loading start\n");
			/* check to make sure the drive isnt locked/loaded already */
			if (vc->locked || vc->loaded)
				return -EBUSY;

			/* just update our virtual cd data */
			if (copy_from_user(&cdemu_info, (void *)arg, sizeof(cdemu_info)))
				return -EFAULT;

			vc->backing_file = fget(cdemu_info.ffd);
			if (!vc->backing_file)
				return -EBADF;

			memcpy(vc->tracks, cdemu_info.tracks, sizeof(cdemu_info.tracks));
			vc->num_tracks = cdemu_info.num_tracks;
			strncpy(vc->comment, cdemu_info.comment, CDEMU_COMMENT_SIZE);
			vc->comment[CDEMU_COMMENT_SIZE-1] = '\0';
			vc->uid = current->uid;
			printk_cdemu(KERN_INFO, "loaded %s cd (%i track[s]) on drive %s by uid %i\n",
			             vc->comment, vc->num_tracks, cdi->name, vc->uid);

			for (tmp = 0; tmp <= vc->num_tracks; tmp++)
			{
				dprintk("cdemu_ioctl: track %d: %02d:%02d:%02d (%lu), offset %lu\n", (tmp == vc->num_tracks) ? CDROM_LEADOUT : tmp + 1,
						vc->tracks[tmp].minute,
						vc->tracks[tmp].second,
						vc->tracks[tmp].frame,
						vc->tracks[tmp].start_sector,
						vc->tracks[tmp].start_offset);
			}

			/* set size of device to image size */
			track = vc->num_tracks;
			vc->size = vc->tracks[track].start_sector * CD_FRAMESIZE;
			gd_sector = vc->size / KERNEL_SECTOR_SIZE;
			set_capacity(vc->gd, gd_sector);
			dprintk("cdemu_ioctl: loading ... size (2048 bytes/sector): %i, sectors (512 bytes/sector): %i\n", vc->size, (int) gd_sector);

			/* only inc use count if there was no cd in the drive */
			if (!vc->loaded)
				UP_MODULE_USE();

			vc->loaded = 1;

			dprintk("cdemu_ioctl: loading end\n");

			return 0;

		case CDEMU_UNLOADCD:
			printk_cdemu(KERN_INFO, "unloaded cd on drive %s\n", cdi->name);
			return unload_cd(vc);

		default:
			dprintk("cdemu_ioctl: Unknown ioctl call 0x%x\n", cmd);
			return -EINVAL;
	}
}

static int cdemu_block_open(struct inode *inode, struct file *file)
{
	dprintk("cdemu_block_open: start\n");

	if (inode && inode->i_bdev && inode->i_bdev->bd_disk && inode->i_bdev->bd_disk->private_data) {
		struct cdemu_device *vc = (struct cdemu_device *)inode->i_bdev->bd_disk->private_data;
		dprintk("cdemu_block_open: going into cdrom_open()\n");
		return cdrom_open(&vc->info, inode, file);
	}

	dprintk("cdemu_block_open: -ENODEV\n");
	return -ENODEV;
}

static int cdemu_block_release(struct inode *inode, struct file *file)
{
	dprintk("cdemu_block_release: start\n");

	if (inode && inode->i_bdev && inode->i_bdev->bd_disk && inode->i_bdev->bd_disk->private_data) {
		struct cdemu_device *vc = (struct cdemu_device *)inode->i_bdev->bd_disk->private_data;
		return cdrom_release(&vc->info, file);
	}

	dprintk("cdemu_block_release: end\n");
	return -ENODEV;
}

static int cdemu_block_ioctl(struct inode *inode, struct file *file, unsigned cmd, unsigned long arg)
{
	dprintk("cdemu_block_ioctl: start\n");

	if (inode && inode->i_bdev && inode->i_bdev->bd_disk && inode->i_bdev->bd_disk->private_data) {
		struct cdemu_device *vc = (struct cdemu_device *)inode->i_bdev->bd_disk->private_data;
    
		int ret = cdemu_ioctl(&vc->info, cmd, arg);
		if (ret != -EINVAL)
			return ret;

		dprintk("cdemu_ioctl: pass ioctl call 0x%x to cdrom_ioctl\n", cmd);      
		ret = cdrom_ioctl(file, &vc->info, inode, cmd, arg);
		dprintk("cdemu_ioctl: ret from cdrom_ioctl 0x%x\n", ret);    
		return ret;
	}

	dprintk("cdemu_block_ioctl: end\n");
	return -ENODEV;
}

static int cdemu_block_media_changed(struct gendisk *disk)
{
	dprintk("cdemu_block_media_changed: start %s\n", disk->disk_name);

	if (disk && disk->private_data) {
		struct cdemu_device *vc = (struct cdemu_device *)disk->private_data;
		dprintk("cdemu_block_media_changed: going into cdrom_media_changed()\n");
		return cdrom_media_changed(&vc->info);
	}

	dprintk("cdemu_block_media_changed: -ENODEV\n");
	return -ENODEV;
}

struct cdrom_device_ops cdemu_dops = {
	.open					= cdemu_open,
	.release				= cdemu_release,
	.drive_status			= cdemu_drive_status,
	.media_changed			= cdemu_media_changed,
	.tray_move				= cdemu_tray_move,
	.lock_door				= cdemu_lock_door,
	.select_speed			= cdemu_select_speed,
	.get_last_session		= cdemu_get_last_session,
	.get_mcn				= NULL,
	.reset					= cdemu_reset,
	.audio_ioctl			= cdemu_audio_ioctl,
	.capability				= CDC_OPEN_TRAY | CDC_LOCK |
							  CDC_MULTI_SESSION | CDC_MEDIA_CHANGED |
							  CDC_PLAY_AUDIO | CDC_GENERIC_PACKET |
							  CDC_DRIVE_STATUS,
	.n_minors				= 1,
	.generic_packet			= cdemu_generic_packet,
};

struct block_device_operations cdemu_bdops = {
	.owner					= THIS_MODULE,
	.open					= cdemu_block_open,
	.release				= cdemu_block_release,
	.ioctl					= cdemu_block_ioctl,
	.media_changed			= cdemu_block_media_changed,
};
