/**
 * File: cdemu.c
 * Authors: Robert Penz <robert.penz@outertech.com>
 *          Mike Frysinger <vapier@gmail.com>
 * Copyright 2002-2004
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the Free Software           *
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.             *
 *                                                                         *
 ***************************************************************************/


#include "cdemu_kernel.h"
#include "cdemu.h"

#define MIN_CDEMUS      1     /* Minimum number of block devices */
#define MAX_CDEMUS      255   /* 2.6 can have more than this but who would use it ? */
#define DEFAULT_CDEMUS  4     /* Default number of block devices */

/* external stuff */
extern struct cdrom_device_ops cdemu_dops;
extern struct block_device_operations cdemu_bdops;
extern void cdemu_request(request_queue_t*);
extern int cdemu_make_request(request_queue_t*, struct bio*);

/* bus functions */
static int virtual_bus_remove(struct device *dev);
static void virtual_bus_shutdown(struct device *dev);
static void virtual_bus_release(struct device *bus);

/* driver functions */
static ssize_t show_driver_version(struct device_driver *drv, char *buf);
static ssize_t show_driver_devices(struct device_driver *drv, char *buf);
static ssize_t store_driver_devices(struct device_driver *drv, const char *buf, size_t count);
static int store_driver_devices_helper(struct device *dev, void *ptr);

/* device functions */
static void cdemu_device_release(struct device *dev);
static struct cdemu_device *add_cdemu_device(int i);
static int remove_cdemu_device_helper(struct device *dev, void *ptr);

static ssize_t show_device_attr(struct device *dev, struct device_attribute *attr, char *buf);

/* global variables */
struct cdemu_driver cdemu_drv = {
	.version = CDEMU_VERSION,
	.major = MAJOR_NR,
	.count = DEFAULT_CDEMUS,
	.drv = {
		.name   = DRIVER_NAME,
		.bus    = &virtual_bus_type,
		.owner  = THIS_MODULE
	}
};

struct bus_type virtual_bus_type = {
	.name     = BUS_NAME,
	.remove   = &virtual_bus_remove,
	.shutdown = &virtual_bus_shutdown
};

struct device virtual_bus = {
	.bus_id  = BUS_NAME,
	.bus     = NULL,
	.parent  = NULL,
	.release = &virtual_bus_release
};

struct devices_vars {
	int n;
	int old, new;
};

static int driver_was_registered = 0;		/* did the driver already exist? */

static DRIVER_ATTR(version, 0444, show_driver_version, NULL);
static DRIVER_ATTR(devices, 0664, show_driver_devices, store_driver_devices);

static DEVICE_ATTR(major, 0444, show_device_attr, NULL);
static DEVICE_ATTR(minor, 0444, show_device_attr, NULL);
static DEVICE_ATTR(loaded, 0444, show_device_attr, NULL);
static DEVICE_ATTR(image, 0444, show_device_attr, NULL);
static DEVICE_ATTR(tracks, 0444, show_device_attr, NULL);



static int virtual_bus_remove(struct device *dev)
{
	if(!dev)
		return -1;

	dprintk("%s: device %s\n", __FUNCTION__, dev->bus_id);

	return 0;
}

static void virtual_bus_shutdown(struct device *dev)
{
	if(!dev)
		return;

	dprintk("%s: device %s\n", __FUNCTION__, dev->bus_id);
}

static void virtual_bus_release(struct device *bus) 
{
	if(!bus)
		return;

	dprintk("%s: bus %s\n", __FUNCTION__, bus->bus_id);
}

static ssize_t show_driver_version(struct device_driver *drv, char *buf)
{
	struct cdemu_driver *cdrv;

	cdrv = to_cdemu_driver(drv);
	if(!drv || !buf || !cdrv)
		return -1;

	return snprintf(buf, PAGE_SIZE, "%s\n", cdrv->version);
}

static ssize_t show_driver_devices(struct device_driver *drv, char *buf)
{
	struct cdemu_driver *cdrv;

	cdrv = to_cdemu_driver(drv);
	if(!drv || !buf || !cdrv)
		return -1;

	return snprintf(buf, PAGE_SIZE, "%i\n", cdrv->count);
}

static int store_driver_devices_helper(struct device *dev, void *ptr)
{
	struct devices_vars *dv;
	
	dv = (struct devices_vars *) ptr;
	if(!dev || !ptr || !dv)
		return -1;

	/* remove devices */
	if(dv->n >= dv->new)
		remove_cdemu_device_helper(dev, NULL);

	dv->n++;

	return 0;
}

static ssize_t store_driver_devices(struct device_driver *drv, const char *buf, size_t count)
{
	char *endp;
	struct devices_vars dv;

	if(!drv || !buf)
		return -1;

	endp = (char *) buf + count - 1;
	dv.old = cdemu_drv.count;
	dv.new = simple_strtol(buf, &endp, 10);

	dprintk("%s: old device count: %i, new device count: %i.\n", __FUNCTION__, dv.old, dv.new);

	if((dv.new < MIN_CDEMUS) || (dv.new > MAX_CDEMUS) || (dv.new == dv.old))
		return (ssize_t) count;

	if(dv.new < dv.old) {
		/* remove devices */
		dv.n = 0;
		device_for_each_child(&virtual_bus, &dv, &store_driver_devices_helper);
	}
	else {
		/* add devices */
		for(dv.n = dv.old; dv.n < dv.new; dv.n++) {
			if(!add_cdemu_device(dv.n))
				return -1;
		}
	}

	cdemu_drv.count = dv.new;

	return (ssize_t) count;
}

static void cdemu_device_release(struct device *dev)
{
	if(!dev)
		return;

	dprintk("%s: device %s\n", __FUNCTION__, dev->bus_id);

	/* Do only LDM related releasing here */
}

static struct cdemu_device *add_cdemu_device(int i)
{
	struct cdemu_device *vc;

	/* allocate device */
	vc = kmalloc(sizeof(struct cdemu_device), GFP_KERNEL);
	if(!vc)
		goto add_device_free;
	memset(vc, 0, sizeof(struct cdemu_device));

	/* init device fields */	
	vc->number = i;
	vc->loaded = 0;
	vc->media_changed = 0;
	vc->locked = 0;
	vc->pause = 0;
	vc->info.ops = &cdemu_dops;
	vc->info.speed = 0; /* FIXME max_speed */
	vc->info.handle = vc;
	vc->info.capacity = 1;
	snprintf(vc->info.name, 20, "%s%i", DEVICE_NAME, i);
	spin_lock_init(&vc->lock);
	dprintk("%s: making cdrom #%i [%p]: %s\n", __FUNCTION__, vc->number, vc, vc->info.name);

	/* register with the linux device model */
	vc->dev.bus = &virtual_bus_type;
	vc->dev.driver = &cdemu_drv.drv; /* driver binding; if set to NULL invokes match & probe */
	vc->dev.release = &cdemu_device_release;
	vc->dev.parent = &virtual_bus;
	strncpy(vc->dev.bus_id, vc->info.name, BUS_ID_SIZE);
	if(device_register(&vc->dev))
		goto add_device_free;
	if(device_create_file(&vc->dev, &dev_attr_major))
		goto add_device_unregister;
	if(device_create_file(&vc->dev, &dev_attr_minor))
		goto add_device_unregister;
	if(device_create_file(&vc->dev, &dev_attr_loaded))
		goto add_device_unregister;
	if(device_create_file(&vc->dev, &dev_attr_image))
		goto add_device_unregister;
	if(device_create_file(&vc->dev, &dev_attr_tracks))
		goto add_device_unregister;

	/* register with the uniform cdrom driver */
	vc->gd = alloc_disk(1);
	if (!vc->gd)
		goto add_device_out;
	vc->gd->major = cdemu_drv.major;
	vc->gd->first_minor = i;
	snprintf(vc->gd->disk_name, 32, "%s%i", DEVICE_NAME, i);
	vc->gd->fops = &cdemu_bdops;
	vc->gd->private_data = vc;
	vc->gd->flags = GENHD_FL_REMOVABLE | GENHD_FL_CD;
	vc->gd->driverfs_dev = &vc->dev;
	set_capacity(vc->gd, 0);

	if(register_cdrom(&vc->info))
		printk(KERN_WARNING "%s: cannot register cdrom\n", __FUNCTION__);

	vc->queue = blk_alloc_queue(GFP_KERNEL);
	blk_queue_make_request(vc->queue, cdemu_make_request);
	if (!vc->queue)
		goto add_device_out;
	vc->gd->queue = vc->queue;
	add_disk(vc->gd);

#ifdef CONFIG_DEVFS_FS
	devfs_mk_bdev(MKDEV(vc->gd->major, vc->gd->first_minor),
	              S_IFBLK|S_IRUSR|S_IWUSR,
	              "cdemu/%i", i);
#endif

	return vc;

add_device_out:
	if (vc->gd) {
		put_disk(vc->gd);
		del_gendisk(vc->gd);
		vc->gd = NULL;
	}
	if (vc->queue) {
		blk_put_queue(vc->queue);
		vc->queue = NULL;
	}

	unregister_cdrom(&vc->info);

add_device_unregister:
	device_unregister(&vc->dev);

add_device_free:
	kfree(vc);

	return NULL;
}

static int remove_cdemu_device_helper(struct device *dev, void *ptr)
{
	struct cdemu_device *vc;

	vc = to_cdemu_device(dev);
	if (!dev || !vc)
		return -1;
		
	dprintk("%s: removing cdrom #%i [%p]: %s\n", __FUNCTION__, vc->number, vc, vc->dev.bus_id);

	if (vc->gd) {
		put_disk(vc->gd);
		del_gendisk(vc->gd);
		vc->gd = NULL;
	}
	if (vc->queue) {
		blk_put_queue(vc->queue);
		vc->queue = NULL;
	}
	
	if (unregister_cdrom(&vc->info))
		printk(KERN_WARNING "%s: cannot unregister cdrom\n", __FUNCTION__);

#ifdef CONFIG_DEVFS_FS
	dprintk("%s: removing devfs\n", __FUNCTION__);
	devfs_remove("cdemu/%i", vc->number);
#endif

	device_unregister(dev);

	kfree(vc);

	return 0;
}

static ssize_t show_device_attr(struct device *dev, struct device_attribute *attr, char *buf)
{
	struct cdemu_device *vc;
	ssize_t             retval;
	
	vc = to_cdemu_device(dev);
	retval = 0;
	if(!dev || !attr || !buf || !vc)
		return 0;

	dprintk("%s: device %s\n", __FUNCTION__,dev->bus_id);
	
	/* choose one of the device attributes */
	if (attr == &dev_attr_major) {
		retval = snprintf(buf, PAGE_SIZE, "%i\n", vc->gd->major);
	}
	else if(attr == &dev_attr_minor) {
		retval = snprintf(buf, PAGE_SIZE, "%i\n", vc->gd->first_minor);
	}
	else if(attr == &dev_attr_loaded) {
		retval = snprintf(buf, PAGE_SIZE, "%i\n", vc->loaded);
	}
	else if(attr == &dev_attr_image) {
		retval = snprintf(buf, PAGE_SIZE, "%s\n", vc->comment);
	}
	else if(attr == &dev_attr_tracks) {
		retval = snprintf(buf, PAGE_SIZE, "%i\n", vc->num_tracks);
	}

	return retval;
}

static int __init cdemu_init(void)
{
	int i, retval;

	dprintk("%s: start\n", __FUNCTION__);

	/* register bus */
	dprintk("%s: registering bus: %s\n", __FUNCTION__, virtual_bus_type.name);
	if(bus_register(&virtual_bus_type))
		goto out;
	if(device_register(&virtual_bus))
		goto out_bus_unregister;

	/* register driver */
	if(!driver_find(DRIVER_NAME, &virtual_bus_type)) {
		dprintk("%s: registering driver: %s\n", __FUNCTION__, cdemu_drv.drv.name);
		driver_was_registered = 0;
		if(driver_register(&cdemu_drv.drv))
			goto out_bus_unregister;
		if(driver_create_file(&cdemu_drv.drv, &driver_attr_version))
			goto out_driver_unregister;
		if(driver_create_file(&cdemu_drv.drv, &driver_attr_devices))
			goto out_driver_unregister;
	}
	else {
		driver_was_registered = 1; /* not possible? */
	}

	/* check if the kernel parameter is correct and register the block devices */
	if ((cdemu_drv.count < MIN_CDEMUS) || (cdemu_drv.count > MAX_CDEMUS)) {
		printk_cdemu(KERN_WARNING, "invalid cdemu_count (must be in range [%i, %i]), using default (%i)\n",
		             MIN_CDEMUS, MAX_CDEMUS, DEFAULT_CDEMUS);
		cdemu_drv.count = DEFAULT_CDEMUS;
	}

	retval = register_blkdev(cdemu_drv.major, DEVICE_NAME);
	if (retval < 0) {
		printk_cdemu(KERN_ERR, "unable to register block device #%i for cdemu device\n", cdemu_drv.major);
		goto out_blkdev;
	} else if (retval)
		cdemu_drv.major = retval;

#if defined(CONFIG_DEVFS_FS)
	devfs_mk_dir("cdemu");
#endif

	for (i = 0; i < cdemu_drv.count; ++i) {
		if(!add_cdemu_device(i))
			goto out_free_devices;
	}

	/* all done report our success */
	dprintk("%s: loaded (%i devices)\n", __FUNCTION__, cdemu_drv.count);
	dprintk("%s: end\n", __FUNCTION__);
	printk_cdemu(KERN_INFO, "cdemu ver. " CDEMU_MODULE_VERSION " loaded.  Registered %i cdemus.\n", cdemu_drv.count); 

	return 0;

out_free_devices:
	device_for_each_child(&virtual_bus, NULL, &remove_cdemu_device_helper);

out_blkdev:
        if (unregister_blkdev(cdemu_drv.major, DEVICE_NAME))
                printk_cdemu(KERN_ERR, "unable to unregister block device\n");

out_driver_unregister:
	driver_unregister(&cdemu_drv.drv);

out_bus_unregister:
	device_unregister(&virtual_bus);
	bus_unregister(&virtual_bus_type);

out:
        printk_cdemu(KERN_ERR, "cdemu_init: ran out of memory\n");

        return -ENOMEM;
}

static void __exit cdemu_exit(void)
{
	dprintk("%s: start\n", __FUNCTION__);

	device_for_each_child(&virtual_bus, NULL, &remove_cdemu_device_helper);

#if defined(CONFIG_DEVFS_FS)
	devfs_remove("cdemu");
#endif

	if (unregister_blkdev(cdemu_drv.major, DEVICE_NAME))
		printk_cdemu(KERN_ERR, "unable to unregister block device\n");

	/* unregister driver */
	if(!driver_was_registered) {
		dprintk("%s: unregistering driver: %s\n", __FUNCTION__, cdemu_drv.drv.name);
		driver_unregister(&cdemu_drv.drv);
	}

	/* unregister bus */
	dprintk("%s: unregistering bus: %s\n", __FUNCTION__, virtual_bus_type.name);
	device_unregister(&virtual_bus);
	bus_unregister(&virtual_bus_type);

	dprintk("%s: end\n", __FUNCTION__);
	printk_cdemu(KERN_INFO, "cdemu ver. " CDEMU_MODULE_VERSION " unloaded.  Unregistered %i cdemus.\n", cdemu_drv.count); 
}

module_init(cdemu_init);
module_exit(cdemu_exit);

#if !defined(MODULE)
static int __init cdemu_count_setup(char *str)
{
	cdemu_drv.count = simple_strtol(str, NULL, 0);
	return 1;
}

__setup("cdemu_count=", cdemu_count_setup);
#endif

MODULE_AUTHOR("Robert Penz <robert.penz@outertech.com>, Mike Frysinger <vapier@gmail.com>");
MODULE_DESCRIPTION("CD Emulator for CD images ver. " CDEMU_MODULE_VERSION);
MODULE_LICENSE("GPL");
module_param_named(cdemu_major, cdemu_drv.major, int, 0444);
module_param_named(cdemu_count, cdemu_drv.count, int, 0444);
MODULE_PARM_DESC(cdemu_major, "Major number to use for devices.");
MODULE_PARM_DESC(cdemu_count, "Number of virtual cdroms must be within the range [1, 255].");
